library IEEE;
use IEEE.std_logic_1164.all;

use work.motion_controller_pkg.all;


entity dac_controller is
    generic (
        channels : natural
    );
    port (
        -- Clock/reset
        clk : IN std_logic;
        strobe_1mhz : IN std_logic;
        reset : IN std_logic;

        -- Requested DAC value
        dac_values : IN dac_vector(channels-1 downto 0);

        -- DAC values to be polled next time
        dac_poll : OUT std_logic;

        -- DAC control systems
        dac_csn : OUT std_logic_vector(channels-1 downto 0) := (others => '1');
        dac_r_wn : OUT std_logic := '0';
        dac_ldac : OUT std_logic := '0';
        dac_rst : OUT std_logic := '0';
        --      Data lines
        db : INOUT std_logic_vector(15 downto 0);
        --      Level shifter lines
        dac_ls_dir : OUT std_logic;
        dac_ls_oen : OUT std_logic_vector(channels-1 downto 0)
    );
end entity dac_controller;

architecture sequencer of dac_controller is
begin
    dac_ls_dir <= '1';
    dac_ls_oen <= (others => '0');

    sequencer_proc:
    process(clk)
        type sequencer_state_t is (reset_state, clear, s1, s2, s3, s4, s5);
        variable sequencer_state : sequencer_state_t := reset_state;
        variable dac_index : natural := 0;

        variable latched_dac_values : dac_vector(channels-1 downto 0);
    begin
        if clk'event and clk = '1' then
            if reset = '1' then
                sequencer_state := reset_state;
                dac_index := 0;

                dac_poll <= '0';
                dac_csn <= (others => '1');
                dac_r_wn <= '0';
                dac_ldac <= '0';
                dac_rst <= '0';
            elsif strobe_1mhz = '1' then
                dac_poll <= '0';
                dac_csn <= (others => '1');
                dac_r_wn <= '0';
                dac_ldac <= '0';
                dac_rst <= '0';

                case sequencer_state is
                    when reset_state =>
                        dac_rst <= '1';
                        sequencer_state := clear;
                        dac_index := 0;
                    when clear =>
                        dac_poll <= '1';
                        sequencer_state := s1;
                    when s1 =>
                        if dac_index = 0 then
                            latched_dac_values := dac_values;
                        end if;

                        db <= std_logic_vector(latched_dac_values(dac_index));

                        sequencer_state := s2;
                    when s2 =>
                        dac_csn(dac_index) <= '0';

                        sequencer_state := s3;
                    when s3 =>
                        -- CSN high
                        dac_index := dac_index + 1;
                        if dac_index = channels then
                            dac_index := 0;
                            sequencer_state := s4;
                        else
                            sequencer_state := s1;
                        end if;
                    when s4 =>
                        dac_ldac <= '1';
                        sequencer_state := s5;
                    when s5 =>
                        dac_poll <= '1';
                        sequencer_state := s1;
                end case;
            end if;
        end if;
    end process sequencer_proc;
end architecture sequencer;
