library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

use work.motion_controller_pkg.all;


entity motion_controller is
    generic (
        channels : natural := 2;
        -- TODO: Set size of vectors based off roll count (or at least a
        -- generic)
        roll_counts : natural_vector
    );
    port (
        -- CPU Interface signals
        clk : IN std_logic;
        reset_n : IN std_logic;

        led : OUT std_logic;

        awid : IN std_logic_vector(11 downto 0);
        awaddr : IN std_logic_vector(11 downto 0);
        awlen : IN std_logic_vector(3 downto 0);
        awsize : IN std_logic_vector(2 downto 0);
        awburst : IN std_logic_vector(1 downto 0);
        awprot : IN std_logic_vector(2 downto 0);
        awvalid : IN std_logic;
        awready : OUT std_logic;

        wid : IN std_logic_vector(11 downto 0);
        wdata : IN std_logic_vector(31 downto 0);
        wstrb : IN std_logic_vector(3 downto 0);
        wlast : IN std_logic;
        wvalid : IN std_logic;
        wready : OUT std_logic;

        bid : OUT std_logic_vector(11 downto 0);
        bresp : OUT std_logic_vector(1 downto 0);
        bvalid : OUT std_logic;
        bready : IN std_logic;

        arid : IN std_logic_vector(11 downto 0);
        araddr : IN std_logic_vector(11 downto 0);
        arlen : IN std_logic_vector(3 downto 0);
        arsize : IN std_logic_vector(2 downto 0);
        arburst : IN std_logic_vector(1 downto 0);
        arlock : IN std_logic_vector(1 downto 0);
        arcache : IN std_logic_vector(3 downto 0);
        arprot : IN std_logic_vector(2 downto 0);
        arvalid : IN std_logic;
        arready : OUT std_logic;

        rid : OUT std_logic_vector(11 downto 0);
        rdata : OUT std_logic_vector(31 downto 0);
        rresp : OUT std_logic_vector(1 downto 0);
        rlast : OUT std_logic;
        rvalid : OUT std_logic;
        rready : IN std_logic;

        -- Motion controller signals
        -- dac controls
        dac_csn : OUT std_logic_vector(channels-1 downto 0) := (others => '1');
        dac_r_wn : OUT std_logic := '0';
        dac_ldac : OUT std_logic := '0';
        dac_rst : OUT std_logic := '0';
        -- data lines to/from DAC (through a level shifter)
        db : INOUT std_logic_vector(15 downto 0);
        -- level shifter control
        dac_ls_dir : OUT std_logic; -- high - towards DAC
        -- high - disable
        dac_ls_oen : OUT std_logic_vector(channels-1 downto 0);

        axis_amp_fault : IN std_logic_vector(channels-1 downto 0);
        axis_forward_limit : IN std_logic_vector(channels-1 downto 0);
        axis_reverse_limit : IN std_logic_vector(channels-1 downto 0);
        axis_home : IN std_logic_vector(channels-1 downto 0);
        axis_encoder_a : IN std_logic_vector(channels-1 downto 0);
        axis_encoder_b : IN std_logic_vector(channels-1 downto 0);
        axis_encoder_z : IN std_logic_vector(channels-1 downto 0);
        axis_inhibit_n : OUT std_logic_vector(channels-1 downto 0)
    );
end entity motion_controller;

architecture structural of motion_controller is
    signal reset : std_logic;
    signal axi_reset : std_logic;

    -- Strobes, generally only high for 1 cycle
    signal reset_strobe : std_logic;

    signal home_strobes : std_logic_vector(channels-1 downto 0);
    signal move_strobes : std_logic_vector(channels-1 downto 0);
    signal stop_strobes : std_logic_vector(channels-1 downto 0);

    -- Registers, value kept until changed
    signal command_directions : std_logic_vector(channels-1 downto 0);
    signal position_commands : position_vector(channels-1 downto 0);

    signal position_actuals : position_vector(channels-1 downto 0);
    signal position_actual_tick_tocks : std_logic_vector(channels-1 downto 0);
    signal quadrature_invalid_counts :
                quadrature_invalid_count_vector(channels-1 downto 0);

    signal dac_values : dac_vector(channels-1 downto 0);
    signal dac_overrides : dac_vector(channels-1 downto 0);
    signal dac_use_overrides : std_logic_vector(channels-1 downto 0);

    signal homes : std_logic_vector(channels-1 downto 0);
    signal negative_limits : std_logic_vector(channels-1 downto 0);
    signal positive_limits : std_logic_vector(channels-1 downto 0);
    signal amp_faults : std_logic_vector(channels-1 downto 0);

    signal axis_direct_inhibits : std_logic_vector(channels-1 downto 0);

    signal done_strobes : std_logic_vector(channels-1 downto 0);
    signal interrupted_strobes : std_logic_vector(channels-1 downto 0);


    signal decoder_home_strobes : std_logic_vector(channels-1 downto 0);

    -- Signals from CPU (AXI bus) to quadrature decoder
    signal axi_quadrature_home_strobes :
            std_logic_vector(channels-1 downto 0);

    -- Signals from motion state machine to quadrature decoder
    signal sm_decoder_home_strobes : std_logic_vector(channels-1 downto 0);

    -- Signals from quadrature decoder to motion state machine
    signal decoder_homed : std_logic_vector(channels-1 downto 0);

    -- Signals from motion state machine to DAC
    signal sm_dac_values : dac_vector(channels-1 downto 0);


    signal strobe_1mhz : std_logic;
begin
    reset <= (not reset_n) or reset_strobe;
    axi_reset <= not reset_n;

    channels_generate:
    for i in channels-1 downto 0 generate
        dac_values(i) <= dac_overrides(i)
                            when dac_use_overrides(i) = '1'
                            else sm_dac_values(i);

        axis_inhibit_n(i) <= not axis_direct_inhibits(i);

        homes(i) <= axis_home(i);
        negative_limits(i) <= axis_reverse_limit(i);
        positive_limits(i) <= axis_forward_limit(i);
        amp_faults(i) <= axis_amp_fault(i);

        quadrature_decoder:
        entity work.quadrature_decoder(behavioral)
            generic map (
                bits => 19,
                glitch_bits => 8,
                roll_count => roll_counts(i),
                home_deglitch_count => 5000,
                debounce_cycle_bits => 6,
                debounce_cycles => to_unsigned(50, 6)
            )
            port map (
                reset => reset,
                clk => clk,
                A => axis_encoder_a(i),
                B => axis_encoder_b(i),
                home => axis_home(i),
                home_strobe => decoder_home_strobes(i),
                homed => decoder_homed(i),
                count => position_actuals(i),
                tick_tock => position_actual_tick_tocks(i),
                glitch_count => quadrature_invalid_counts(i)
            );

        state_machine:
        entity work.motion_controller_state_machine(state_machine)
            port map (
                clk => clk,
                reset => reset,
                position_command => position_commands(i),
                command_direction => command_directions(i),
                position_actual => position_actuals(i),
                tick_tock => position_actual_tick_tocks(i),
                home_strobe => home_strobes(i),
                move_strobe => move_strobes(i),
                stop_strobe => stop_strobes(i),
                decoder_homed => decoder_homed(i),
                negative_limit => negative_limits(i),
                positive_limit => positive_limits(i),
                dac_value => sm_dac_values(i),
                decoder_home_strobe => sm_decoder_home_strobes(i),
                done_strobe => done_strobes(i),
                interrupted_strobe => interrupted_strobes(i)
            );

        -- OR the state machine probes once they exist
        decoder_home_strobes(i) <= axi_quadrature_home_strobes(i) or
                                    sm_decoder_home_strobes(i);
    end generate channels_generate;

    cpu_interface: entity work.motion_controller_axi_interface(state_machines)
        generic map (
            channels => channels
        )
        port map (
            clk => clk,
            reset => axi_reset,
            led => led,

            awid => awid,
            awaddr => awaddr,
            awlen => awlen,
            awsize => awsize,
            awburst => awburst,
            awprot => awprot,
            awvalid => awvalid,
            awready => awready,

            wid => wid,
            wdata => wdata,
            wstrb => wstrb,
            wlast => wlast,
            wvalid => wvalid,
            wready => wready,

            bid => bid,
            bresp => bresp,
            bvalid => bvalid,
            bready => bready,

            arid => arid,
            araddr => araddr,
            arlen => arlen,
            arsize => arsize,
            arburst => arburst,
            arlock => arlock,
            arcache => arcache,
            arprot => arprot,
            arvalid => arvalid,
            arready => arready,

            rid => rid,
            rdata => rdata,
            rresp => rresp,
            rlast => rlast,
            rvalid => rvalid,
            rready => rready,

            -- Strobes, generally only high for 1 cycle
            reset_strobe => reset_strobe,

            home_strobes => home_strobes,
            move_strobes => move_strobes,
            stop_strobes => stop_strobes,

            -- Registers, value kept until changed
            command_directions => command_directions,
            position_commands => position_commands,

            position_actuals => position_actuals,
            position_actual_tick_tocks => position_actual_tick_tocks,
            quadrature_invalid_counts => quadrature_invalid_counts,

            dac_actuals => dac_values,
            dac_request => dac_overrides,
            dac_asserts => dac_use_overrides,

            quadrature_home_strobes => axi_quadrature_home_strobes,

            -- Direct inputs
            homes => homes,
            negative_limits => negative_limits,
            positive_limits => positive_limits,
            amp_faults => amp_faults,

            -- Direct outputs (or through AND/OR)
            axis_inhibits => axis_direct_inhibits,

            -- Input strobes
            done_strobes => done_strobes,
            interrupted_strobes => interrupted_strobes
        );

        strobe_1mhz_gen:
        process(clk)
            variable count : unsigned(5 downto 0) := (others => '0');
        begin
            if clk'event and clk = '1' then
                count := count + 1;

                if count = to_unsigned(50, 6) then
                    count := to_unsigned(0, 6);
                    strobe_1mhz <= '1';
                else
                    strobe_1mhz <= '0';
                end if;
            end if;
        end process strobe_1mhz_gen;


        dac_interface: entity work.dac_controller(sequencer)
            generic map (
                channels => channels
            )
            port map (
                clk => clk,
                strobe_1mhz => strobe_1mhz,
                reset => reset,
                dac_values => dac_values,
                dac_csn => dac_csn,
                dac_r_wn => dac_r_wn,
                dac_ldac => dac_ldac,
                dac_rst => dac_rst,
                db => db,
                dac_ls_dir => dac_ls_dir,
                dac_ls_oen => dac_ls_oen
            );
end architecture structural;
