library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

use work.motion_controller_pkg.all;


-- Memory map
-- 0x0 - Axis x commands/status
--    0 - command_direction - write 1 to move in the positive direction, 0 for
--                          negative
--    1 - home_strobe - write 1 to strobe home, if idle or in the reset done
--                          state, home the positioner in the direction in
--                          command_direction, if not in idle/reset done,
--                          causes an error busy
--    2 - move_strobe - write 1 to strobe move, if idle, move the positioner in
--                          the direction of command_direction, if no in idle
--                          causes an error busy
--    3 - stop_strobe - write 1 to strobe stop, if homing or moving stop and
--                          go back to reset or idle state, set
--                          interrupted_flag, not an error if not moving (just
--                          a NOP in that case, also no interrupted_flag)
--    4 - done_flag - read - if a move/home as completed
--    5 - clear_done_flag - write - write 1 to clear done_flag
--    6 - interrupted_flag - read - if a move/home was interrupted for any
--                          reason
--    7 - clear_interrupted_flag - write - write 1 to clear interrupted_flag
--    8 - error_flag - read - if an error has occured
--    9-15 - error_reason - read what error has occurred
--    16 - clear_error_flag - write - write 1 to clear error_flag
--    17-23 - state - read - what the status of the positioner is, like moving,
--                          homing, idle, etc
--    24-31 - invalid_quadrature_count - read - count of invalid quadrature
--                          transitions (like 00 -> 11)
-- 0x4 - Axis x position_target/actual
--    18-0 - position_command - write - the position command for the next move
--                          strobe
--    18-0 - position_actual - read - the actual position of the positioner
--    19 - position_tick_tock_actual - read - the tick/tock value, changes each
--                          revolution, used to track crosses across 359.999/0
--                          unambiguously
-- 0x8 - Axis x debug
--    15-0 - dac_value - Read - what actual DAC value is going out
--                       Write - DAC force value, used if bit 16 is set
--    16 - dac_force - Read/Write - 0 - normal value, 1 - force dac_value out
--    17 - home - Read - is home switch depressed currently
--    18 - negative_limit - Read - is the negative limit switch depressed
--              currently
--    19 - positive_limit - Read - is the positive limit switch depressed
--              currently
--    20 - amp_fault - Read - is the amp fault asserted
--    21 - axis_inhibit - Read/Write - 0 - do nothing, 1 - inhibit the axis
--              This value is ORed with any other inhibits.
--    22 - home_strobe - Write - write 1 to strobe the home on the quadrature
--              decoder
-- 100 -
--    0 - reset_strobe - write 1 to reset this peripheral (implicitly
--                          everything else is ignored)
-- 104 - LED strobe, write a number to keep the LED for that many clocks

entity motion_controller_axi_interface is
    generic (
        channels : natural := 2
    );
    port (
        -- CPU Interface signals
        clk : IN std_logic;
        reset : IN std_logic;

        led : OUT std_logic;

        awid : IN std_logic_vector(11 downto 0);
        awaddr : IN std_logic_vector(11 downto 0);
        awlen : IN std_logic_vector(3 downto 0);
        awsize : IN std_logic_vector(2 downto 0);
        awburst : IN std_logic_vector(1 downto 0);
        awprot : IN std_logic_vector(2 downto 0);
        awvalid : IN std_logic;
        awready : OUT std_logic;

        wid : IN std_logic_vector(11 downto 0);
        wdata : IN std_logic_vector(31 downto 0);
        wstrb : IN std_logic_vector(3 downto 0);
        wlast : IN std_logic;
        wvalid : IN std_logic;
        wready : OUT std_logic;

        bid : OUT std_logic_vector(11 downto 0);
        bresp : OUT std_logic_vector(1 downto 0);
        bvalid : OUT std_logic;
        bready : IN std_logic;

        arid : IN std_logic_vector(11 downto 0);
        araddr : IN std_logic_vector(11 downto 0);
        arlen : IN std_logic_vector(3 downto 0);
        arsize : IN std_logic_vector(2 downto 0);
        arburst : IN std_logic_vector(1 downto 0);
        arlock : IN std_logic_vector(1 downto 0);
        arcache : IN std_logic_vector(3 downto 0);
        arprot : IN std_logic_vector(2 downto 0);
        arvalid : IN std_logic;
        arready : OUT std_logic;

        rid : OUT std_logic_vector(11 downto 0);
        rdata : OUT std_logic_vector(31 downto 0);
        rresp : OUT std_logic_vector(1 downto 0);
        rlast : OUT std_logic;
        rvalid : OUT std_logic;
        rready : IN std_logic;

        -- Strobes, generally only high for 1 cycle
        reset_strobe : OUT std_logic;

        home_strobes : OUT std_logic_vector(channels-1 downto 0);
        move_strobes : OUT std_logic_vector(channels-1 downto 0);
        stop_strobes : OUT std_logic_vector(channels-1 downto 0);

        -- Registers, value kept until changed
        command_directions : OUT std_logic_vector(channels-1 downto 0);
        position_commands : OUT position_vector(channels-1 downto 0);

        position_actuals : IN position_vector(channels-1 downto 0);
        position_actual_tick_tocks : IN std_logic_vector(channels-1 downto 0);
        quadrature_invalid_counts : IN
                quadrature_invalid_count_vector(channels-1 downto 0);

        dac_actuals : IN dac_vector(channels-1 downto 0);
        dac_request : OUT dac_vector(channels-1 downto 0);
        dac_asserts : BUFFER std_logic_vector(channels-1 downto 0);

        quadrature_home_strobes : OUT std_logic_vector(channels-1 downto 0);

        -- Direct inputs
        homes : IN std_logic_vector(channels-1 downto 0);
        negative_limits : IN std_logic_vector(channels-1 downto 0);
        positive_limits : IN std_logic_vector(channels-1 downto 0);
        amp_faults : IN std_logic_vector(channels-1 downto 0);

        -- Direct outputs (or through AND/OR)
        axis_inhibits : BUFFER std_logic_vector(channels-1 downto 0);

        -- Input strobes
        done_strobes : IN std_logic_vector(channels-1 downto 0);
        interrupted_strobes : IN std_logic_vector(channels-1 downto 0)
    );
end entity motion_controller_axi_interface;

architecture state_machines of motion_controller_axi_interface is
    -- Status registers
    type error_t is (no_error, invalid_command);
    type error_vector is array (natural range <>) of error_t;
    signal errors : error_vector(channels-1 downto 0);

    signal led_strobe : std_logic;
    signal led_count_s : std_logic_vector(31 downto 0);

    -- Flags
    signal done_flags : std_logic_vector(channels-1 downto 0);
    signal interrupted_flags : std_logic_vector(channels-1 downto 0);
begin
    bresp <= (others => '0');

    led_proc:
    process(clk)
        variable led_count_v : unsigned(31 downto 0);
    begin
        if clk'event and clk='1' then
            if reset = '1' then
                led_count_v := (others => '0');
                led <= '0';
            else
                if led_strobe = '1' and led_count_v = to_unsigned(0, 32) then
                    led_count_v := unsigned(led_count_s);
                end if;

                if led_count_v > to_unsigned(0, 1) then
                    led <= '1';
                    led_count_v := led_count_v - to_unsigned(1, 1);
                else
                    led <= '0';
                end if;
            end if;
        end if;
    end process led_proc;

    write_proc:
    process(clk)
        variable b_wait : boolean;
        variable awid_reg : std_logic_vector(11 downto 0);
        variable awaddr_reg : std_logic_vector(11 downto 0);

        variable do_write : boolean := false;
        variable channel_id : natural;
    begin
        if clk'event and clk='1' then
            if reset = '1' then
                awready <= '0';
                wready <= '0';
                bid <= (others => '0');
                bvalid <= '0';
                b_wait := false;

                do_write := false;

                led_count_s <= (others => '0');
                led_strobe <= '0';

                -- Strobes init to 0 (esp the reset strobe)
                reset_strobe <= '0';
                home_strobes <= (others => '0');
                move_strobes <= (others => '0');
                stop_strobes <= (others => '0');

                -- Registers
                command_directions <= (others => '1');
                position_commands <= (others => (others => '0'));
                dac_request <= (others => "1000000000000000");
                dac_asserts <= (others => '0');
                axis_inhibits <= (others => '0');

                -- Status registers
                for channel_id_rst in channels-1 downto 0 loop
                    errors(channel_id_rst) <= no_error;
                end loop;

                -- TODO: How to handle these wrt axi vs peripheral reset
                done_flags <= (others => '0');
                interrupted_flags <= (others => '0');
            else
                do_write := false;

                for i in channels-1 downto 0 loop
                    if done_strobes(i) = '1' then
                        done_flags(i) <= '1';
                    end if;

                    if interrupted_strobes(i) = '1' then
                        interrupted_flags(i) <= '1';
                    end if;
                end loop;

                awready <= '0';
                wready <= '0';

                led_strobe <= '0';

                -- Clear one-shot strobes
                reset_strobe <= '0';
                home_strobes <= (others => '0');
                move_strobes <= (others => '0');
                stop_strobes <= (others => '0');

                quadrature_home_strobes <= (others => '0');

                if b_wait and bready = '1' then
                    bvalid <= '0';
                    b_wait := false;
                end if;

                if awvalid = '1' and wvalid = '1' and not b_wait then
                    do_write := true;
                    awaddr_reg := awaddr;

                    awready <= '1';
                    wready <= '1';
                    bvalid <= '1';
                    bid <= awid;
                    b_wait := true;
                end if;
            end if;

            if do_write then
                case awaddr_reg(11 downto 8) is
                    when "0000" =>
                        channel_id := natural(to_integer(unsigned(
                                            awaddr_reg(7 downto 4))));
                        case awaddr_reg(3 downto 0) is
                            when "0000" =>
                                command_directions(channel_id) <= wdata(0);

                                case std_logic_vector'(wdata(1), wdata(2),
                                                       wdata(3)) is
                                    when "000" =>
                                        null;
                                    when "100" =>
                                        home_strobes(channel_id) <= '1';
                                    when "010" =>
                                        move_strobes(channel_id) <= '1';
                                    when "001" =>
                                        stop_strobes(channel_id) <= '1';
                                    when others =>
                                        errors(channel_id) <= invalid_command;
                                end case;

                                if wdata(5) = '1' then
                                    done_flags(channel_id) <= '0';
                                end if;

                                if wdata(7) = '1' then
                                    interrupted_flags(channel_id) <= '0';
                                end if;

                                if wdata(16) = '1' then
                                    errors(channel_id) <= no_error;
                                end if;
                            when "0100" =>
                                position_commands(channel_id) <=
                                        unsigned(wdata(18 downto 0));
                            when "1000" =>
                                dac_request(channel_id) <=
                                        unsigned(wdata(15 downto 0));
                                dac_asserts(channel_id) <= wdata(16);
                                axis_inhibits(channel_id) <= wdata(21);
                                quadrature_home_strobes(channel_id) <=
                                        wdata(22);
                            when others =>
                                null;
                        end case;
                    when "0001" =>
                        case awaddr_reg(7 downto 0) is
                            when "00000000" =>
                                reset_strobe <= wdata(0);
                            when "00000100" =>
                                led_strobe <= '1';
                                led_count_s <= wdata;
                            when others =>
                                null;
                        end case;
                    when others =>
                        null;
                end case;
            end if;
        end if;
    end process write_proc;

    rresp <= (others => '0');

    read_proc:
    process(clk)
        variable arid_reg : std_logic_vector(11 downto 0);
        variable araddr_reg : std_logic_vector(11 downto 0);
        variable arlen_reg : unsigned(3 downto 0);
        variable arsize_reg : std_logic_vector(2 downto 0);
        variable arburst_reg : std_logic_vector(1 downto 0);

        type state_t is (idle, transferring);
        variable state_v : state_t := idle;

        variable waste_count : integer;

        variable update_rdata : boolean := false;

        variable channel_id : natural;
    begin
        if clk'event and clk='1' then
            update_rdata := false;

            if reset = '1' then
                arready <= '1';
                rid <= (others => '0');
                rlast <= '0';
                rvalid <= '0';
                state_v := idle;
            else
                case state_v is 
                    when idle =>
                        rvalid <= '0';
                        rlast <= '0';
                        if arvalid = '1' then
                            arready <= '0';
                            arid_reg := arid;
                            araddr_reg := araddr;
                            arlen_reg := unsigned(arlen);
                            arsize_reg := arsize;
                            arburst_reg := arburst;
                            state_v := transferring;

                            -- rvalid <= '1';
                            -- rid <= arid_reg;

                            -- rdata <= (4 => '1', others => '0');
                            -- rdata(3 downto 0) <= std_logic_vector(arlen_reg);

                            -- if arlen_reg = to_unsigned(0, 4) then
                            --     rlast <= '1';
                            -- else
                            --     rlast <= '0';
                            -- end if;
                        else
                            arready <= '1';
                        end if;
                    when transferring =>
                        rvalid <= '1';
                        rid <= arid_reg;

                        update_rdata := true;

                        if rready = '1' then
                            if arlen_reg = to_unsigned(0, 4) then
                                state_v := idle;
                                rvalid <= '0';
                                rlast <= '0';
                            else
                                arlen_reg := arlen_reg - to_unsigned(1, 4);

                                rdata <= (4 => '1', others => '0');
                                rdata(3 downto 0) <=
                                    std_logic_vector(arlen_reg);

                                if arlen_reg = to_unsigned(0, 4) then
                                    rlast <= '1';
                                else
                                    rlast <= '0';
                                end if;
                            end if;
                        end if;
                end case;
            end if;

            if update_rdata then
                case araddr_reg(11 downto 8) is
                    when "0000" =>
                        channel_id := natural(to_integer(unsigned(
                                            araddr_reg(7 downto 4))));
                        case araddr_reg(3 downto 0) is
                            when "0000" =>
                                rdata <= (4 => done_flags(channel_id),
                                          6 => interrupted_flags(channel_id),
                                          8 => '0', -- TODO: Flag error
                                          -- TODO: Encode errors
                                          9 to 15 => '0',
                                          -- TODO: Encode state
                                          17 to 23 => '0',
                                          others => '0'
                                      );
                                rdata(31 downto 24) <=
                                        std_logic_vector(
                                            quadrature_invalid_counts(
                                                channel_id));
                            when "0100" =>
                                rdata <= (19 => position_actual_tick_tocks(
                                                channel_id),
                                          others => '0');
                                rdata(18 downto 0) <= std_logic_vector(
                                                position_actuals(channel_id));
                            when "1000" =>
                                rdata <= (16 => dac_asserts(channel_id),
                                          17 => homes(channel_id),
                                          18 => negative_limits(channel_id),
                                          19 => positive_limits(channel_id),
                                          20 => amp_faults(channel_id),
                                          21 => axis_inhibits(channel_id),
                                          others => '0');
                                rdata(15 downto 0) <= std_logic_vector(
                                        dac_actuals(channel_id));
                            when others =>
                                rdata <= std_logic_vector(
                                            to_unsigned(1337, 32));
                        end case;
                    when "0001" =>
                        case araddr_reg(7 downto 0) is
                            when "00000000" =>
                                rdata <= (others => '0');
                            when "00000100" =>
                                rdata <= (others => '0');
                            when others =>
                                rdata <= std_logic_vector(
                                            to_unsigned(1337, 32));
                        end case;
                    when others =>
                        rdata <= std_logic_vector(to_unsigned(42, 32));
                end case;
            end if;
        end if;
    end process read_proc;
end architecture state_machines;
