library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity quadrature_decoder is
    generic (
        bits : natural;
        glitch_bits : natural;
        roll_count : natural;
        -- Angle have to turn before trust home is fresh
        home_deglitch_count : natural;
        debounce_cycle_bits : natural;
        debounce_cycles : unsigned
    );
    port (
        reset : IN std_logic; -- An already syncronous reset
        clk : IN std_logic;
        A, B : IN std_logic;
        home : IN std_logic;

        -- Initiate a homing
        home_strobe : IN std_logic;
        -- Home has been found
        homed : OUT std_logic;

        count : OUT unsigned(bits-1 downto 0);
        tick_tock : OUT std_logic;
        glitch_count : OUT unsigned(glitch_bits-1 downto 0);
        -- Synchronized and debounced version of home. count will be set to 0
        -- the clock cycle after the rising edge of this signal
        debounced_home : BUFFER std_logic
    );
end entity quadrature_decoder;

architecture behavioral of quadrature_decoder is
    signal up, down, glitch : std_logic;

    type input_record is record
        A : std_logic;
        B : std_logic;
        home : std_logic;
    end record;

    signal syncd_inputs : input_record;
    signal syncd_reset : std_logic;
begin
    -- Synchronize to the clock domain using a few extra flip-flops.
    input_synchronizer_process:
    process(clk)
        constant length : integer := 3;

        type input_record_vector is array (natural range <>) of input_record;

        variable sync_vec : input_record_vector(length-1 downto 0);
        variable reset_vec : std_logic_vector(length downto 0)
            := (others => '1');
    begin
        if clk'event and clk='1' then
            if reset='1' then
                reset_vec := (others => '1');
            end if;

            syncd_inputs <= sync_vec(0);
            syncd_reset <= reset_vec(0);

            --crossed_A <= A_vec(0);
            --crossed_B <= B_vec(0);
            --crossed_home <= home_vec(0);
            for i in 0 to length-2 loop
                sync_vec(i) := sync_vec(i+1);
                --A_vec(i) := A_vec(i+1);
                --B_vec(i) := B_vec(i+1);
                --home_vec(i) := home_vec(i+1);
                reset_vec(i) := reset_vec(i+1);
            end loop;

            sync_vec(length-1).A := A;
            sync_vec(length-1).B := B;
            sync_vec(length-1).home := home;
            reset_vec(length-1) := reset_vec(length);
            reset_vec(length) := '0';
        end if;
    end process input_synchronizer_process;

    -- debounce:
    -- aaaabbabbb -> aaaabbbbbb - swallow the a within the b, do not swallow
    --      the bb
    -- Assume that first edges coming along are true (i.e. that if there is an
    --      electrical transition then there is a mechanical one, just that
    --      multiple electrical transitions do not imply more mechanical ones).
    -- When edge comes along, it is immediately accepted, then all future edges
    --      are delayed until the signal has not changed for x cycles.
    debounce_home:
    process(clk)
        variable old_home : std_logic;
        variable last_edge_count : unsigned(debounce_cycle_bits-1 downto 0);
    begin
        if clk'event and clk = '1' then
            if syncd_reset = '1' then
                last_edge_count := debounce_cycles;
                debounced_home <= syncd_inputs.home;
                old_home := syncd_inputs.home;
            else
                -- If a change has occured
                if syncd_inputs.home /= old_home then
                    -- If we're not in the midst of changes
                    --if last_edge_count = to_unsigned(0,
                    --                                 debounce_cycle_bits) then
                    --    debounced_home <= syncd_inputs.home;
                    --end if;
                    -- In any case refresh the count
                    last_edge_count := debounce_cycles;
                -- If no change has occured
                else
                    -- If not yet steady
                    if last_edge_count > to_unsigned(0,
                                                     debounce_cycle_bits) then
                        last_edge_count := last_edge_count - 1;
                    -- If steady
                    else
                        debounced_home <= syncd_inputs.home;
                    end if;
                end if;

                old_home := syncd_inputs.home;
            end if;
        end if;
    end process debounce_home;


    -- 00 ^ up
    -- 10
    -- 11
    -- 01
    quadrature_process:
    process(clk, syncd_inputs)
        variable AB_vec, old_AB_vec : std_logic_vector(1 downto 0);
        variable glitched : boolean := false;
    begin
        if clk'event and clk='1' then
            AB_vec(0) := syncd_inputs.A;
            AB_vec(1) := syncd_inputs.B;

            if reset = '1' then
                glitched := false;
                old_AB_vec := AB_vec;
            else
                if AB_vec = not old_AB_vec then
                    if not glitched then
                        glitch <= '1';
                        glitched := true;
                    else
                        glitch <= '0';
                    end if;
                else
                    glitch <= '0';
                    glitched := false;

                    up <= '0';
                    down <= '0';

                    case old_AB_vec is
                        when "00" =>
                            case AB_vec is
                                when "01" => up <= '1';
                                when "10" => down <= '1';
                                when "11" => null; --glitch <= '1';
                                when others => null;
                            end case;
                        when "01" =>
                            case AB_vec is
                                when "11" => up <= '1';
                                when "00" => down <= '1';
                                when "10" => null; --glitch <= '1';
                                when others => null;
                            end case;
                        when "11" =>
                            case AB_vec is
                                when "10" => up <= '1';
                                when "01" => down <= '1';
                                when "00" => null; --glitch <= '1';
                                when others => null;
                            end case;
                        when "10" =>
                            case AB_vec is
                                when "00" => up <= '1';
                                when "11" => down <= '1';
                                when "01" => null; -- glitch <= '1';
                                when others => null;
                            end case;
                        when others => glitch <= '1';
                    end case;

                    old_AB_vec := AB_vec;
                end if;
            end if;
        end if;
    end process quadrature_process;

    count_process:
    process(clk)
        variable homed_v : boolean;
        variable homing_v : boolean;

        variable count_var : unsigned(bits-1 downto 0);
        variable tick_tock_v : std_logic;
        variable glitch_count_var : unsigned(glitch_bits-1 downto 0);
        variable old_debounced_home : std_logic;
    begin
        if clk'event and clk='1' then
            if syncd_reset = '1' then
                count_var := (others => '0');
                glitch_count_var := (others => '0');
                old_debounced_home := debounced_home;
                homed_v := false;
                homing_v := false;
                tick_tock_v := '0';
            else
                if home_strobe = '1' then
                    homing_v := true;
                    homed_v := false;
                    count_var := to_unsigned(0, bits);
                    tick_tock_v := '0';
                end if;

                if up = '1' then
                    if count_var = to_unsigned(roll_count-1, bits) then
                        tick_tock_v := not tick_tock_v;
                        count_var := to_unsigned(0, bits);
                    else
                        count_var := count_var + to_unsigned(1,2);
                    end if;
                elsif down = '1' then
                    if count_var = to_unsigned(0, bits) then
                        tick_tock_v := not tick_tock_v;
                        count_var := to_unsigned(roll_count-1, bits);
                    else
                        count_var := count_var - to_unsigned(1,2);
                    end if;
                end if;

                if homing_v and debounced_home = '0' then
                    if ((count_var > to_unsigned(home_deglitch_count, bits) and
                         tick_tock_v = '0') or
                        (count_var < to_unsigned(
                            roll_count-home_deglitch_count, bits)
                         and tick_tock_v = '1')) then
                        homing_v := false;
                        homed_v := true;
                    end if;
                    count_var := to_unsigned(0, bits);
                    tick_tock_v := '0';
                end if;

                if glitch = '1' then
                    glitch_count_var := glitch_count_var + to_unsigned(1, 1);
                end if;
            end if;

            count <= count_var;
            glitch_count <= glitch_count_var;

            tick_tock <= tick_tock_v;
            if homed_v then
                homed <= '1';
            else
                homed <= '0';
            end if;
        end if;
    end process count_process;
end architecture behavioral;
