library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity simple_axi_slave is
    port (
        clk : IN std_logic;
        reset_n : IN std_logic;

        led : OUT std_logic;

        awid : IN std_logic_vector(11 downto 0);
        awaddr : IN std_logic_vector(11 downto 0);
        awlen : IN std_logic_vector(3 downto 0);
        awsize : IN std_logic_vector(2 downto 0);
        awburst : IN std_logic_vector(1 downto 0);
        awprot : IN std_logic_vector(2 downto 0);
        awvalid : IN std_logic;
        awready : OUT std_logic;

        wid : IN std_logic_vector(11 downto 0);
        wdata : IN std_logic_vector(31 downto 0);
        wstrb : IN std_logic_vector(3 downto 0);
        wlast : IN std_logic;
        wvalid : IN std_logic;
        wready : OUT std_logic;

        bid : OUT std_logic_vector(11 downto 0);
        bresp : OUT std_logic_vector(1 downto 0);
        bvalid : OUT std_logic;
        bready : IN std_logic;

        arid : IN std_logic_vector(11 downto 0);
        araddr : IN std_logic_vector(11 downto 0);
        arlen : IN std_logic_vector(3 downto 0);
        arsize : IN std_logic_vector(2 downto 0);
        arburst : IN std_logic_vector(1 downto 0);
        arlock : IN std_logic_vector(1 downto 0);
        arcache : IN std_logic_vector(3 downto 0);
        arprot : IN std_logic_vector(2 downto 0);
        arvalid : IN std_logic;
        arready : OUT std_logic;

        rid : OUT std_logic_vector(11 downto 0);
        rdata : OUT std_logic_vector(31 downto 0);
        rresp : OUT std_logic_vector(1 downto 0);
        rlast : OUT std_logic;
        rvalid : OUT std_logic;
        rready : IN std_logic
    );
end entity simple_axi_slave;

architecture simple of simple_axi_slave is
    signal reset : std_logic;

    signal led_count_s : std_logic_vector(31 downto 0);
    signal led_strobe : std_logic;
begin
    reset <= not reset_n;

    bresp <= (others => '0');

    led_proc:
    process(clk)
        variable led_count_v : unsigned(31 downto 0);
    begin
        if clk'event and clk='1' then
            if reset = '1' then
                led_count_v := (others => '0');
                led <= '0';
            else
                if led_strobe = '1' and led_count_v = to_unsigned(0, 32) then
                    led_count_v := unsigned(led_count_s);
                end if;

                if led_count_v > to_unsigned(0, 1) then
                    led <= '1';
                    led_count_v := led_count_v - to_unsigned(1, 1);
                else
                    led <= '0';
                end if;
            end if;
        end if;
    end process led_proc;

    write_proc:
    process(clk)
        variable b_wait : boolean;
        variable awid_reg : std_logic_vector(11 downto 0);
    begin
        if clk'event and clk='1' then
            if reset = '1' then
                awready <= '0';
                wready <= '0';
                bid <= (others => '0');
                bvalid <= '0';
                b_wait := false;

                led_count_s <= (others => '0');
                led_strobe <= '0';
            else
                awready <= '0';
                wready <= '0';

                led_strobe <= '0';

                if b_wait and bready = '1' then
                    bvalid <= '0';
                    b_wait := false;
                end if;

                if awvalid = '1' and wvalid = '1' and not b_wait then
                    if awaddr = "000000000100" then
                        led_count_s <= wdata;
                        led_strobe <= '1';
                    end if;

                    awready <= '1';
                    wready <= '1';
                    bvalid <= '1';
                    bid <= awid;
                    b_wait := true;
                end if;
            end if;
        end if;
    end process write_proc;

    rresp <= (others => '0');

    read_proc:
    process(clk)
        variable arid_reg : std_logic_vector(11 downto 0);
        variable araddr_reg : std_logic_vector(11 downto 0);
        variable arlen_reg : unsigned(3 downto 0);
        variable arsize_reg : std_logic_vector(2 downto 0);
        variable arburst_reg : std_logic_vector(1 downto 0);

        type state_t is (idle, transferring);
        variable state_v : state_t := idle;

        variable waste_count : integer;
    begin
        if clk'event and clk='1' then
            if reset = '1' then
                arready <= '1';
                rid <= (others => '0');
                rlast <= '0';
                rvalid <= '0';
                state_v := idle;
            else
                case state_v is 
                    when idle =>
                        rvalid <= '0';
                        rlast <= '0';
                        if arvalid = '1' then
                            arready <= '0';
                            arid_reg := arid;
                            arlen_reg := unsigned(arlen);
                            arsize_reg := arsize;
                            arburst_reg := arburst;
                            state_v := transferring;

                            rvalid <= '1';
                            rid <= arid_reg;

                            rdata <= (4 => '1', others => '0');
                            rdata(3 downto 0) <= std_logic_vector(arlen_reg);

                            if arlen_reg = to_unsigned(0, 4) then
                                rlast <= '1';
                            else
                                rlast <= '0';
                            end if;
                        else
                            arready <= '1';
                        end if;
                    when transferring =>
                        rvalid <= '1';
                        rid <= arid_reg;

                        if rready = '1' then
                            if arlen_reg = to_unsigned(0, 4) then
                                state_v := idle;
                                rvalid <= '0';
                                rlast <= '0';
                            else
                                arlen_reg := arlen_reg - to_unsigned(1, 4);

                                rdata <= (4 => '1', others => '0');
                                rdata(3 downto 0) <= std_logic_vector(arlen_reg);

                                if arlen_reg = to_unsigned(0, 4) then
                                    rlast <= '1';
                                else
                                    rlast <= '0';
                                end if;
                            end if;
                        end if;
                end case;
            end if;
        end if;
    end process read_proc;
end architecture simple;
